/* Kernel Programming */
#define LINUX

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <asm/io.h>

#include <linux/seq_file.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>

#include <linux/spinlock_types.h>
#include <linux/delay.h>

/******************************************************************************/
/* Constants								      */
/******************************************************************************/
#define MAX_CPUS							      20
#define MAX_CLOCK_STR_LENGTH						      32
#define MAX_SAMPLE_ENTRIES						10000000
#define INITIAL_CPU_ARRAY						       \
		{ -1, -1, -1, -1, -1, -1, -1, -1,			       \
		  -1, -1, -1, -1, -1, -1, -1, -1,			       \
		  -1, -1, -1, -1}
#define INITIAL_WEIGHT_ARRAY						\
		{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 			\
		  0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
/******************************************************************************/
/* Entries								      */
/******************************************************************************/
enum {
	LB_MIGRATE_SAMPLE = 0,
	IMBALANCE_CALC_SAMPLE,
};

typedef struct sample_entry {
	unsigned long long sched_clock;
	char entry_type;
	union {
		struct {
			int src_cpu, target_cpu, pid;
			unsigned long load, imbalance;
		} lb_migrate_sample;
		struct {
			unsigned long local_load, busiest_load, sds_load;
			unsigned long imbalance;
		} imbalance_calc_sample;
	} data;
} sample_entry_t;

sample_entry_t *sample_entries = NULL;
unsigned long current_sample_entry_id = 0;
unsigned long n_sample_entries = 0;

/******************************************************************************/
/* Hook function types							      */
/******************************************************************************/
typedef void (* sp_record_lb_migrate_t)(int, int, unsigned long, unsigned long);
typedef void (* sp_record_imbalance_calc_t)(unsigned long, unsigned long, unsigned long, unsigned long);

// typedef void (*record_rq_resv_size_t)(int, int);
// typedef void (* sp_record_ipc_t)(int, int, long, long);

/******************************************************************************/
/* Prototypes								      */
/******************************************************************************/
extern void set_sp_module_record_lb_migrate(sp_record_lb_migrate_t __sp_module_record_lb_migrate);
extern void set_sp_module_record_imbalance_calc(sp_record_imbalance_calc_t __sp_module_record_imbalance_calc);

// extern void set_module_record_rq_resv_size
// 	(record_rq_resv_size_t __module_record_rq_resv_size);
// extern void set_module_record_ipc
// 	(sp_record_ipc_t __module_record_ipc);

static void rq_stats_do_work(struct seq_file *m, unsigned long iteration);

/******************************************************************************/
/* Functions								      */
/******************************************************************************/
static void *my_seq_start(struct seq_file *s, loff_t *pos)
{
	static unsigned long iteration = 0;
	if (*pos == 0 || *pos < n_sample_entries) {
		return &iteration;
	} else {
		*pos = 0;
		return NULL;
	}
}

static void *my_seq_next(struct seq_file *s, void *v, loff_t *pos)
{

	unsigned long *iteration = (unsigned long *)v;

	(*iteration)++;
	(*pos)++;

	if (*iteration < n_sample_entries)
		return v;
	else
		return NULL;
}

static void my_seq_stop(struct seq_file *s, void *v) {}

static int my_seq_show(struct seq_file *s, void *v)
{
	rq_stats_do_work(s, *(unsigned long *)v);
	return 0;
}

static struct seq_operations my_seq_ops = {
	.start = my_seq_start,
	.next  = my_seq_next,
	.stop  = my_seq_stop,
	.show  = my_seq_show
};

static int rq_stats_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &my_seq_ops);
}

static const struct proc_ops rq_stats_fops = {
	// .proc_owner   = THIS_MODULE,
	.proc_open	= rq_stats_open,
	.proc_read	= seq_read,
	.proc_lseek  = seq_lseek,
	.proc_release = seq_release,
};

static void rq_stats_do_work(struct seq_file *m, unsigned long iteration)
{
	sample_entry_t *entry;
	set_sp_module_record_lb_migrate(NULL);
	set_sp_module_record_imbalance_calc(NULL);

	n_sample_entries = current_sample_entry_id < MAX_SAMPLE_ENTRIES ?
			   current_sample_entry_id : MAX_SAMPLE_ENTRIES;

	entry = &sample_entries[iteration];


	seq_printf(m, "	 %llu nsecs", entry->sched_clock);

	if (entry->entry_type == LB_MIGRATE_SAMPLE)
	{
		seq_printf(m, " LB %4d %4d %lu %lu", entry->data.lb_migrate_sample.src_cpu,
				entry->data.lb_migrate_sample.target_cpu,
				entry->data.lb_migrate_sample.load,
				entry->data.lb_migrate_sample.imbalance);
	} else if (entry->entry_type == IMBALANCE_CALC_SAMPLE)
	{
		seq_printf(m, " IC %lu %lu %lu %lu", entry->data.imbalance_calc_sample.busiest_load,
				entry->data.imbalance_calc_sample.sds_load,
				entry->data.imbalance_calc_sample.local_load,
				entry->data.imbalance_calc_sample.imbalance);
	}
	
	seq_printf(m, "\n");
}


void sched_profiler_record_lb_migrate(int src_cpu, int target_cpu,
                unsigned long task_load, unsigned long imbalance)
{
        unsigned long __current_sample_entry_id;

        __current_sample_entry_id =
                __sync_fetch_and_add(&current_sample_entry_id, 1);

        if (__current_sample_entry_id >= MAX_SAMPLE_ENTRIES)
        {
                return;
        }

        sample_entries[__current_sample_entry_id]
                .sched_clock = ktime_get_mono_fast_ns();
        sample_entries[__current_sample_entry_id]
                .entry_type = LB_MIGRATE_SAMPLE;
        sample_entries[__current_sample_entry_id].data.lb_migrate_sample.src_cpu = src_cpu;
        sample_entries[__current_sample_entry_id].data.lb_migrate_sample.target_cpu = target_cpu;
	sample_entries[__current_sample_entry_id].data.lb_migrate_sample.load = task_load;
        sample_entries[__current_sample_entry_id].data.lb_migrate_sample.imbalance = imbalance;
}

void sched_profiler_record_imbalance_calc(unsigned long busiest_avg_load, unsigned long sds_avg_load,
					  unsigned long local_avg_load, unsigned long imbalance)
{
	unsigned long __current_sample_entry_id;

        __current_sample_entry_id =
                __sync_fetch_and_add(&current_sample_entry_id, 1);

        if (__current_sample_entry_id >= MAX_SAMPLE_ENTRIES)
        {
                return;
        }

        sample_entries[__current_sample_entry_id]
                .sched_clock = ktime_get_mono_fast_ns();
        sample_entries[__current_sample_entry_id]
                .entry_type = IMBALANCE_CALC_SAMPLE;
	sample_entries[__current_sample_entry_id].data.imbalance_calc_sample.busiest_load = busiest_avg_load;
	sample_entries[__current_sample_entry_id].data.imbalance_calc_sample.sds_load = sds_avg_load;
	sample_entries[__current_sample_entry_id].data.imbalance_calc_sample.local_load = local_avg_load;
	sample_entries[__current_sample_entry_id].data.imbalance_calc_sample.imbalance = imbalance;
}


static int __init my_init(void)
{
	unsigned long sample_entries_size =
		sizeof(sample_entry_t) * MAX_SAMPLE_ENTRIES;

	proc_create("sched_profiler", S_IRUGO, NULL, &rq_stats_fops);
	printk("sched_profiler: kmalloc fail (bytes)\n");
	if(!(sample_entries = vmalloc(sample_entries_size)))
	{
		printk("sched_profiler: kmalloc fail (%lu bytes)\n",
			   sample_entries_size);
		remove_proc_entry("sched_profile", NULL);

		return -ENOMEM;
	}

	set_sp_module_record_lb_migrate(sched_profiler_record_lb_migrate);
	set_sp_module_record_imbalance_calc(sched_profiler_record_imbalance_calc);
	return 0;
}

static void __exit my_cleanup(void)
{
	set_sp_module_record_lb_migrate(NULL);
	set_sp_module_record_imbalance_calc(NULL);
	ssleep(1);

	vfree(sample_entries);
	remove_proc_entry("sched_profiler", NULL);
}


module_init(my_init);
module_exit(my_cleanup);

MODULE_LICENSE("GPL");

